#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: LGPL-3.0-or-later
#

from mido.ports import BaseOutput
from typing import Callable, Type

from giradischi.backends.base import MidiOutputBackendBase

class MidoBackend:
	get_devices: Callable[[], list[dict]]
	Output: Type[BaseOutput]

class MidoBackendOutputBackendBase(MidiOutputBackendBase):
	backend: Type[MidoBackend]
	api: str

	def __init__(self) -> None:
		super().__init__()

		self.device = None

		devices = self.get_devices()
		if devices:
			self.device = devices[0]

	@classmethod
	def is_available(cls):
		return cls.backend is not None

	def can_have_multiple_devices(self):
		return True

	def get_devices(self):
		devices = self.backend.get_devices()

		return [device["name"] for device in devices if device["is_output"]]

	def get_device(self):
		return self.device

	def set_device(self, device: str):
		devices = self.get_devices()

		assert device in devices, "Device not found"

		self.device = device

	def open_device(self, **kwargs):
		assert self.device, "No device set"

		return self.backend.Output(self.device, **self._add_api(kwargs))

	# Utils
	def _add_api(self, kwargs):
		"""Adds the API to the keyword arguments."""
		if getattr(self, 'api', None) and 'api' not in kwargs:
			kwargs['api'] = self.api
		return kwargs



#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: LGPL-3.0-or-later
#

try:
	from giradischi.backends.fluidsynth import mido_backend
except (ImportError, ModuleNotFoundError):
	mido_backend = None

from giradischi.backends.mido_backend_base import MidoBackendOutputBackendBase

class FluidSynthOutputBackend(MidoBackendOutputBackendBase):
	backend = mido_backend

	@classmethod
	def get_name(cls):
		return "FluidSynth"

	def is_virtual(self):
		return True



#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: LGPL-3.0-or-later
#
"""giradischi GUI."""

from PySide6.QtCore import QFile, Signal
from PySide6.QtGui import QAction
from PySide6.QtUiTools import QUiLoader
from PySide6.QtWidgets import (
	QApplication,
	QFileDialog,
	QLabel,
	QListWidget,
	QListWidgetItem,
	QProgressBar,
	QPushButton,
	QStatusBar,
	QStyle,
)
from pathlib import Path
from threading import Thread
from time import sleep

from giradischi.backends import get_backend_by_name, get_available_backends
from giradischi.utils.midi_player import MidiPlayer

ui_path = Path(__file__).parent

class GiradischiUI(QApplication):
	update_time = Signal(float)

	def __init__(self, file: Path = None) -> None:
		super().__init__()
		self.setApplicationName("giradischi")
		self.setApplicationDisplayName("Giradischi")

		mainwindow_ui_file = QFile(ui_path / "mainwindow.ui")
		mainwindow_ui_file.open(QFile.ReadOnly)

		backend_selector_dialog_ui_file = QFile(ui_path / "backend_selector_dialog.ui")
		backend_selector_dialog_ui_file.open(QFile.ReadOnly)

		backend_settings_dialog_ui_file = QFile(ui_path / "backend_settings_dialog.ui")
		backend_settings_dialog_ui_file.open(QFile.ReadOnly)

		loader = QUiLoader()
		self.mainwindow = loader.load(mainwindow_ui_file)
		self.backend_selector_dialog = loader.load(backend_selector_dialog_ui_file)
		self.backend_settings_dialog = loader.load(backend_settings_dialog_ui_file)

		# mainwindow
		self.status_bar: QStatusBar = self.mainwindow.findChild(QStatusBar, "statusBar")

		self.open_file_action: QAction = self.mainwindow.findChild(QAction, "openFileAction")
		self.open_file_action.triggered.connect(self._select_file)

		self.open_file_button: QPushButton = self.mainwindow.findChild(QPushButton, "openFileButton")
		self.open_file_button.setIcon(self.style().standardIcon(QStyle.SP_DirIcon))
		self.open_file_button.clicked.connect(self._select_file)

		self.play_button: QPushButton = self.mainwindow.findChild(QPushButton, "playPauseButton")
		self.play_button.setIcon(self.style().standardIcon(QStyle.SP_MediaPlay))
		self.play_button.clicked.connect(self._play_pause)

		self.stop_button: QPushButton = self.mainwindow.findChild(QPushButton, "stopButton")
		self.stop_button.setIcon(self.style().standardIcon(QStyle.SP_MediaStop))
		self.stop_button.clicked.connect(self._stop)

		self.settings_button: QPushButton = self.mainwindow.findChild(QPushButton, "settingsButton")
		self.settings_button.setIcon(self.style().standardIcon(QStyle.SP_MediaVolume))
		self.settings_button.clicked.connect(self._open_backend_selector_dialog)

		self.title_label: QLabel = self.mainwindow.findChild(QLabel, "titleLabel")
		self.current_time_label: QLabel = self.mainwindow.findChild(QLabel, "currentTimeLabel")
		self.duration_time_label: QLabel = self.mainwindow.findChild(QLabel, "durationTimeLabel")
		self.progress_bar: QProgressBar = self.mainwindow.findChild(QProgressBar, "progressBar")

		# backend_selector_dialog
		self.backend_selector_list_widget: QListWidget = self.backend_selector_dialog.findChild(QListWidget, "backendSelectorListWidget")
		self.backend_selector_list_widget.itemClicked.connect(self._change_backend)

		self.backend_settings_button: QPushButton = self.backend_selector_dialog.findChild(QPushButton, "backendSettingsButton")
		self.backend_settings_button.clicked.connect(self._open_backend_settings_dialog)

		# backend_settings_dialog
		self.backend_settings_list_widget: QListWidget = self.backend_settings_dialog.findChild(QListWidget, "devicesListWidget")
		self.backend_settings_list_widget.itemClicked.connect(self._change_device)

		self.midi_player = MidiPlayer()
		self.opened_file: Path = None

		self.update_time.connect(self._update_time_label)

		self.thread = Thread(target=self._daemon, daemon=True)
		self.thread.start()

		if file and file.suffix == ".mid":
			self._open_file(file)

	def start_ui(self):
		self.mainwindow.show()

		self.exec()

	def _update_time_label(self, time: float):
		if self.midi_player.is_playing():
			self.current_time_label.setText(self._format_time(time))
			self.progress_bar.setValue(time / self.midi_player.file.length * 100)
		else:
			self.current_time_label.setText("00:00")
			self.progress_bar.setValue(0)

	def _daemon(self):
		while True:
			self.update_time.emit(self.midi_player.current_time)
			sleep(1)

	def _select_file(self) -> None:
		filename, _ = QFileDialog.getOpenFileName(self.mainwindow,
				"Select MIDI file", "", "MIDI files (*.mid)")
		if not filename:
			return

		self._open_file(Path(filename))

	def _open_file(self, file: Path):
		self.opened_file = file
		self.midi_player.open_file(self.opened_file)
		self.title_label.setText(self.opened_file.name)
		self.duration_time_label.setText(self._format_time(self.midi_player.file.length))

	def _play_pause(self) -> None:
		self.midi_player.toggle()
		icon = QStyle.SP_MediaPause if self.midi_player.is_playing() else QStyle.SP_MediaPlay
		self.play_button.setIcon(self.style().standardIcon(icon))

	def _stop(self):
		self.midi_player.stop()

	def _format_time(self, time: float) -> str:
		minutes = int(time // 60)
		seconds = int(time % 60)
		return f"{minutes:02}:{seconds:02}"

	def _open_backend_selector_dialog(self):
		backends = [backend.get_name() for backend in get_available_backends()]
		self.backend_selector_list_widget.clear()
		self.backend_selector_list_widget.addItems(backends)

		if self.midi_player.backend:
			current_backend = self.midi_player.backend.get_name()
			if current_backend in backends:
				self.backend_selector_list_widget.setCurrentRow(backends.index(current_backend))

		self.backend_selector_dialog.show()

	def _open_backend_settings_dialog(self):
		devices = self.midi_player.backend.get_devices()
		self.backend_settings_list_widget.clear()
		self.backend_settings_list_widget.addItems(devices)

		if self.midi_player.backend:
			current_device = self.midi_player.backend.get_device()
			if current_device in devices:
				self.backend_settings_list_widget.setCurrentRow(devices.index(current_device))

		self.backend_settings_dialog.show()

	def _change_backend(self, item: QListWidgetItem):
		new_backend_name = item.text()
		if self.midi_player.backend and (new_backend_name == self.midi_player.backend.get_name()):
			return

		try:
			backend = get_backend_by_name(new_backend_name)()
		except Exception as e:
			self.status_bar.showMessage(f"Error: {e}")
			return

		try:
			self.midi_player.set_backend(backend)
		except Exception as e:
			self.status_bar.showMessage(f"Error: {e}")

	def _change_device(self, item: QListWidgetItem):
		self.midi_player.backend.set_device(item.text())



#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: LGPL-3.0-or-later
#

from fluidsynth import Synth
from mido.ports import BaseOutput

def get_devices():
	return [
		{
			'name': 'default',
			'is_input': False,
			'is_output': True,
		},
	]

# Apparently we can't create attributes on BaseOutput classes
_synths: dict[int, Synth] = {}
_sfids: dict[int, int] = {}

class Output(BaseOutput):
	def _open(self, **kwargs):
		self_id = id(self)

		_synths[self_id] = Synth()
		synth = _synths[self_id]

		# TODO: Add Soundfont GUI settings
		synth.setting("audio.driver", "pulseaudio")
		synth.start()
		_sfids[self_id] = synth.sfload("/usr/share/sounds/sf2/FluidR3_GM.sf2", 1)

	def _close(self):
		self_id = id(self)

		if not self_id in _synths:
			return

		synth = _synths[self_id]

		if self_id in _sfids:
			synth.sfunload(_sfids[self_id], 1)
			del _sfids[self_id]

		synth.delete()
		del _synths[self_id]

	def _send(self, message):
		self_id = id(self)

		if self_id not in _synths:
			return

		synth = _synths[self_id]

		if message.type == 'note_on':
			synth.noteon(message.channel, message.note, message.velocity)
		elif message.type == 'note_off':
			synth.noteoff(message.channel, message.note)
		elif message.type == 'control_change':
			synth.cc(message.channel, message.control, message.value)
		elif message.type == 'pitchwheel':
			synth.pitch_bend(message.channel, message.pitch)
		elif message.type == 'program_change':
			synth.program_change(message.channel, message.program)
		else:
			print(f"Unimplemented message type: {message.type}")



#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: LGPL-3.0-or-later
#

from mido.ports import BaseOutput

class MidiOutputBackendBase:
	"""MIDI output backend."""
	def __init__(self):
		"""Initialize the MIDI output backend."""
		if not self.is_available():
			raise RuntimeError("Backend not available")

	@classmethod
	def get_name(cls) -> str:
		"""Returns the name of the backend."""
		raise NotImplementedError

	@classmethod
	def is_available(cls) -> bool:
		"""Returns True if the backend is available on the system."""
		raise NotImplementedError

	def is_virtual(self) -> bool:
		"""Returns True if the device is virtual."""
		raise NotImplementedError

	def can_have_multiple_devices(self) -> bool:
		"""Returns True if the backend can have multiple devices."""
		raise NotImplementedError

	def get_devices(self) -> list[str]:
		"""Returns a list of MIDI output devices available on the system.

		If only one device is always available (like for virtual synthesizers), either return
		the name of the device or a dummy one (like "default" or "virtual").
		If no device is available, return an empty list.
		"""
		raise NotImplementedError

	def get_device(self) -> str:
		"""Returns the current MIDI device.

		If can_have_multiple_devices() returns False,
		return the device returned by get_devices()."""
		raise NotImplementedError

	def set_device(self, device: str) -> None:
		"""Sets the MIDI device to use.

		Value should be one of the values returned by get_devices().

		If can_have_multiple_devices() returns False, don't implement this method."""
		raise NotImplementedError

	def open_device(self, **kwargs) -> BaseOutput:
		"""Opens the MIDI device.

		Returns a BaseOutput instance.
		Assert that the device is set before calling this function.
		"""
		raise NotImplementedError



#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: LGPL-3.0-or-later
#
"""giradischi MIDI output backends"""

from typing import Type

from giradischi.backends.base import MidiOutputBackendBase

# Backends
from giradischi.backends.alsamidi import AlsaMidiOutputBackend
from giradischi.backends.fluidsynth import FluidSynthOutputBackend
from giradischi.backends.portmidi import PortMidiOutputBackend
from giradischi.backends.rtmidi import RtMidiOutputBackend

BACKENDS: list[Type[MidiOutputBackendBase]] = [
	AlsaMidiOutputBackend,
	FluidSynthOutputBackend,
	PortMidiOutputBackend,
	RtMidiOutputBackend,
]

DEFAULT_BACKEND = RtMidiOutputBackend

def get_available_backends() -> list[Type[MidiOutputBackendBase]]:
	"""Get a list of available backends."""
	return [backend for backend in BACKENDS if backend.is_available()]

def get_backend_by_name(name: str):
	"""Find a backend by name."""
	for backend in BACKENDS:
		if backend.get_name() == name:
			return backend

	raise ValueError(f"Backend {name} not found")



#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: LGPL-3.0-or-later
#

from sebaubuntu_libs.liblogging import setup_logging

from giradischi.ui import GiradischiUI

def main():
	setup_logging()

	GiradischiUI().start_ui()



#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: LGPL-3.0-or-later
#

try:
	from alsa_midi import mido_backend
except ModuleNotFoundError:
	mido_backend = None

from platform import system

from giradischi.backends.mido_backend_base import MidoBackendOutputBackendBase

class AlsaMidiOutputBackend(MidoBackendOutputBackendBase):
	backend = mido_backend

	@classmethod
	def get_name(cls):
		return "ALSA MIDI"

	@classmethod
	def is_available(cls):
		return super().is_available() is not None and system() == "Linux"

	def is_virtual(self):
		return False



#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: LGPL-3.0-or-later
#
"""giradischi module."""

from pathlib import Path

__version__ = "1.0.0"

module_path = Path(__file__).parent
current_path = Path.cwd()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['giradischi',
 'giradischi.backends',
 'giradischi.backends.alsamidi',
 'giradischi.backends.fluidsynth',
 'giradischi.backends.portmidi',
 'giradischi.backends.rtmidi',
 'giradischi.ui',
 'giradischi.utils']

package_data = \
{'': ['*']}

install_requires = \
['PySide6>=6.3.0,<7.0.0', 'mido>=1.2.10,<2.0.0']

setup_kwargs = {
    'name': 'giradischi',
    'version': '1.0.0',
    'description': 'GUI MIDI player supporting multiple backends',
    'long_description': '# giradischi\n\n[![PyPi version](https://img.shields.io/pypi/v/giradischi)](https://pypi.org/project/giradischi/)\n\nGUI MIDI player supporting multiple backends\n\nRequires Python 3.9\n\n## Installation\n\n```sh\npip3 install giradischi\n```\n\n### Backends\n\n-   ALSA MIDI: Install [alsa-midi](https://pypi.org/project/alsa-midi/) with `pip3 install alsa-midi` and follow the instructions provided [here](https://python-alsa-midi.readthedocs.io/en/latest/overview.html#installation)\n-   FluidSynth: Install [pyFluidSynth](https://pypi.org/project/pyFluidSynth/) with `pip3 install pyFluidSynth` and follow the instructions provided [here](https://github.com/nwhitehead/pyfluidsynth#requirements)\n-   PortMidi: Follow the instructions provided [here](https://mido.readthedocs.io/en/latest/backends/portmidi.html)\n-   RtMidi: Install [python-rtmidi](https://pypi.org/project/python-rtmidi) with `pip3 install python-rtmidi` or install [rtmidi-python](https://pypi.org/project/rtmidi-python) with `pip3 install rtmidi-python`\n\n## How to use\n\n```sh\npython3 -m giradischi\n```\n\n```\n#\n# Copyright (C) 2022 Sebastiano Barezzi\n#\n# SPDX-License-Identifier: LGPL-3.0-or-later\n#\n```\n',
    'author': 'Sebastiano Barezzi',
    'author_email': 'barezzisebastiano@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/SebaUbuntu/giradischi',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.11',
}


setup(**setup_kwargs)



#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: LGPL-3.0-or-later
#

from mido import MidiFile
from pathlib import Path
from threading import Thread, Event
from time import sleep, time

from giradischi.backends import DEFAULT_BACKEND
from giradischi.backends.base import MidiOutputBackendBase

class MidiPlayer:
	"""A basic MIDI player, with play/pause/stop functionality."""
	def __init__(self, backend: MidiOutputBackendBase = None) -> None:
		"""Create a new MIDI player."""
		self.backend = backend if backend else DEFAULT_BACKEND()
		self.file: MidiFile = None

		self.is_started = False

		self.play_event = Event()
		self.stop_event = Event()

		self.current_time = 0.0

		self.thread = Thread(target=self._daemon, daemon=True)
		self.thread.start()

	def set_backend(self, backend: MidiOutputBackendBase) -> None:
		if self.is_started:
			raise RuntimeError("Cannot change backend while playing, stop first")

		self.backend = backend

	def is_playing(self) -> bool:
		"""Return True if the player is currently playing."""
		return self.is_started and self.play_event.is_set()

	def is_paused(self) -> bool:
		"""Return True if the player is currently paused but a file is loaded."""
		return self.is_started and not self.is_playing()

	def is_stopped(self) -> bool:
		"""Return True if the player is currently stopped and no file is loaded."""
		return not self.is_started

	def _daemon(self) -> None:
		while True:
			self.play_event.wait()
			self.stop_event.clear()
			with self.backend.open_device() as output:
				self.is_started = True

				start_time = time()
				self.current_time = 0.0

				for msg in self.file:
					self.current_time += msg.time

					playback_time = time() - start_time
					duration_to_next_event = self.current_time - playback_time

					if duration_to_next_event > 0.0:
						sleep(duration_to_next_event)

					if msg.is_meta:
						continue

					if self.stop_event.is_set():
						output.reset()
						break

					if self.play_event.is_set():
						output.send(msg)
					else:
						# Panic if the user has paused the playback
						output.panic()
						self.play_event.wait()

				self.is_started = False

			self.current_time = 0.0
			self.play_event.clear()
			self.stop_event.clear()

	def open_file(self, midi_file: Path) -> None:
		"""Stops the current playing file if needed and loads a new one."""
		self.stop()
		self.file = MidiFile(midi_file)

	def play(self) -> None:
		"""Start playing the current file."""
		if self.is_playing():
			return

		if not self.file:
			return

		self.play_event.set()

	def pause(self) -> None:
		"""Pause the current playing file."""
		if self.is_paused():
			return

		self.play_event.clear()

	def toggle(self) -> None:
		"""Either play or pause the current file."""
		if self.is_playing():
			self.pause()
		else:
			self.play()

	def stop(self) -> None:
		"""Stop the current playing file and unload it from the daemon."""
		if self.is_stopped():
			return

		self.stop_event.set()
		# Let the thread resume and handle the stop event
		self.play_event.set()



#
# Copyright (C) 2022 Sebastiano Barezzi
#
# SPDX-License-Identifier: LGPL-3.0-or-later
#

try:
	from mido.backends import portmidi
except OSError:
	portmidi = None

from giradischi.backends.mido_backend_base import MidoBackendOutputBackendBase

class PortMidiOutputBackend(MidoBackendOutputBackendBase):
	backend = portmidi

	@classmethod
	def get_name(cls):
		return "PortMidi"

	def is_virtual(self):
		return False
